from .reminder_task import ReminderTask
from ..utils.datafiles import text_template

class UnclosedReminder(ReminderTask):
    REMINDER_TYPE = "Issue assigned but not closed"
    DEFAULT_TEMPLATE_FILE = text_template("unclosed.txt")
    CONFIG = "unclosed-reminder"

    def _get_relevant_issues(self):
        for issue in self.bot.get_open_issues():
            if issue.assignees:
                yield issue

    def _extract_date(self, issue, config):
        return issue.date_last_assigned


if __name__ == "__main__":
    UnclosedReminder.run_cli()



import logging
import warnings
_logger = logging.getLogger(__name__)

from .task import Task

class AssignmentToGMail(Task):
    CONFIG = 'assignment-to-gmail'

    @staticmethod
    def make_parser():
        parser = Task.make_parser()
        parser.add_argument("issue_number", type=int)
        return parser

    @classmethod
    def inject_config_parameters(cls, config, opts):
        my_config = config['workflows'].get(cls.CONFIG)
        if my_config:
            my_config["issue_number"] = opts.issue_number

        return config, opts

    def _build_config(self):
        return self.config

    def _run(self, config, dry):
        issue_number = config['issue_number']
        _logger.info(f"LOADING ISSUE {issue_number}")
        issue = self.bot.get_issue(issue_number)
        if not issue.is_ticket_issue:
            _logger.info("EXITING: No email ticket associated with "
                         f"issue {issue_number}")
            return

        assignees = set(issue.assignees)
        ghuser_to_gmail_label = {
            mem.github: mem.label for mem in self.team
        }

        # figure out who isn't on the team
        non_team = assignees - set(ghuser_to_gmail_label)
        if non_team:
            warnings.warn("No GMail labels for the following assignees: "
                          f"{non_team}")

        labels_to_assign = [
            ghuser_to_gmail_label[user] for user in assignees - non_team
        ]
        _logger.info(f"FOR TICKET '{issue.unique_id}' ASSIGNING LABELS "
                     f"{labels_to_assign}")
        if not dry:
            self.inbox.set_labels(issue.unique_id, labels_to_assign)


if __name__ == "__main__":
    AssignmentToGMail.run_cli()



import string
from datetime import datetime, timedelta

import yaml

import logging
_logger = logging.getLogger(__name__)

from .task import Task

class ReminderTask(Task):
    REMINDER_TYPE = None
    DEFAULT_TEMPLATE_FILE = None

    def _build_config(self):
        config = dict(self.config)
        config['delay'] = timedelta(**config['delay'])
        template_file_name = config.get('template',
                                        self.DEFAULT_TEMPLATE_FILE)
        with open(template_file_name, 'r') as f:
            config['template'] = string.Template(f.read())

        return config

    def craft_reminder_comment(self, template, notify):
        frontmatter = yaml.dump({"REMINDER": self.REMINDER_TYPE})
        if notify:
            notification = " ".join(f"@{user}" for user in notify)
            notification += "\n\n"
        else:
            notification = ""

        main_content = template.substitute()  # TODO: add things here

        body = frontmatter + "\n---\n\n" + notification + main_content
        return body

    def _get_relevant_issues(self):
        raise NotImplementedError()

    def _extract_date(self, issue, config):
        raise NotImplementedError()

    def _single_issue_check(self, issue, config, latest_date, dry):
        _logger.debug(f"CHECKING ISSUE {issue.number}")
        if self._extract_date(issue, config) < latest_date:
            _logger.info(f"CREATING COMMENT FOR ISSUE {issue.number}")
            comment = self.craft_reminder_comment(
                template=config['template'],
                notify=config['notify']
            )
            _logger.info("COMMENT CONTENTS:\n" + comment)
            if not dry:
                self.bot.make_comment(issue.number, comment)

    def get_relevant_issues(self):
        issues = self._get_relevant_issues()
        if self.config['email-ticket-only']:
            issues = (iss for iss in issues if iss.is_ticket_issue)
        return issues

    def _run(self, config, dry):
        latest_date = datetime.now() - config['delay']
        for issue in self.get_relevant_issues():
            self._single_issue_check(issue, config, latest_date, dry)



import itertools
import string
from datetime import timedelta, datetime
from email.mime.text import MIMEText


import logging
_logger = logging.getLogger(__name__)

from ..issues import Issue, NoIssue
from .task import Task
from ..utils.datafiles import text_template


# FILTERS
#
# Filters are designed to identify emails that should not be made into GitHub issues.
# The filter returns a Callable[[Message], bool]. This callable should return True if
# the Message should be excluded from the messages to turn into issues.

def message_from_bot(inbox, bot, team, config):
    def bot_filter_inner(msg):
        return bot.smtp and bot.smtp.user in msg.get("From")

    return bot_filter_inner

def message_from_team_not_to(inbox, bot, team, config):
    def team_filter_inner(msg):
        is_team = any(mem.email in msg.get("From") for mem in team)
        is_to = inbox.user in msg.get("To")
        return is_team and not is_to

    return team_filter_inner


def _reply_template(config):
    default_filename = text_template("email_reply.txt")
    template_filename = config.get('template', default_filename)
    with open(template_filename) as f:
        template_str = f.read()
    reply_template = string.Template(template_str)

    return reply_template


class EmailsToIssues(Task):
    """Task to create GitHub issue from emails.
    """
    CONFIG = "emails-to-issues"
    FILTERS = {
        'bot': message_from_bot,
        'team': message_from_team_not_to,
    }

    def _build_filters(self, filt_config):
        filters = []
        for filt in filt_config:
            if not filt.get("active", True):
                continue
            filter_builder = self.FILTERS[filt['name']]
            new_filter = filter_builder(self.inbox, self.bot, self.team, filt)
            filters.append(new_filter)
        return filters

    def _build_config(self):
        config_dict = self.config
        filter_config = config_dict.get("filters")
        filters = self._build_filters(filter_config)
        recent = timedelta(**config_dict.get("recent", {'hours': 48}))

        config = {
            'filters': filters,
            'recent': recent,
        }

        if reply_config := config_dict.get("reply-inbox"):
            if reply_config.get("active", True):
                config.update({'reply_template': _reply_template(reply_config)})

        return config

    def single_email_to_issue(self, msg, dry):
        # create the issue
        contents = Issue.issue_body_from_message(msg)
        _logger.info(f"CREATING ISSUE\ntitle: {msg.subject}:\n{contents}")

        if not dry:
            issue = self.bot.create_issue(msg.subject, contents)
            _logger.info("CREATED ISSUE {issue.number}")
        else:
            # used in dry run only
            issue = NoIssue()

        return issue

    def send_reply_email(self, orig, issue, reply_template, dry):
        content = reply_template.substitute(
            GITHUB_URL=issue.html_url,
        )
        subj = orig.subject
        if not subj[:4].upper() == "RE: ":
            subj = "RE: " + subj

        reply_email = MIMEText(content)
        reply_email["Subject"] = subj
        reply_email["To"] = self.inbox.user
        reply_email["From"] = self.bot.smtp.user
        reply_email["In-Reply-To"] = orig.get("Message-ID")
        if thread_topic := orig.get("Thread-Topic"):  # 🖕 Microsoft
            reply_email["Thread-Topic"] = thread_topic

        _logger.info(f"SENDING EMAIL:\n{reply_email.as_string()}")
        if not dry:
            self.bot.smtp.sendmail(reply_email, [self.inbox.user])

    def _run(self, config, dry):
        _logger.debug(f"CONFIG: {config}")
        since = datetime.now() - config['recent']
        emails = self.inbox.get_emails(since=since)
        filtered = itertools.dropwhile(
            lambda x: any(f(x) for f in config['filters']),
            emails
        )

        issues = self.bot.get_all_email_ticket_issues()
        id_to_message = {msg.unique_id: msg for msg in filtered}
        id_to_issue = {iss.unique_id: iss for iss in issues}

        ids_to_add = set(id_to_message) - set(id_to_issue)
        for id_ in ids_to_add:
            msg = id_to_message[id_]
            issue = self.single_email_to_issue(msg, dry)
            if template := config.get('reply_template'):
                if not self.bot.smtp:
                    # TODO: fail faster on this; validate config
                    raise RuntimeError(
                        "SMTP must be defined for bot to send replies."
                    )

                self.send_reply_email(msg, issue, template, dry)

if __name__ == "__main__":
    EmailsToIssues.run_cli()



from importlib.resources import files


def workflow_template(filename):
    return files('ticgithub.data.workflows') / filename


def text_template(filename):
    return files('ticgithub.data.templates') / filename


DEFAULT_BUILD_CONFIG = files('ticgithub.data') / "build_config.yml"



import yaml
from datetime import datetime

import github

class NonTicketIssueError(Exception):
    pass

class Issue:
    """
    Simple wrapper around github.Issue to define our supported API.
    """
    def __init__(self, issue: github.Issue):
        self._issue = issue

    @staticmethod
    def issue_body_from_message(message):
        frontmatter = {
            'ticket_id': message.unique_id,
            'From': message.get("From"),
            'Date': str(message.date),
            'Subject': message.subject,
        }
        yaml_part = yaml.dump(frontmatter)
        return yaml_part + "\n---\n" + message.get_content()

    @staticmethod
    def _get_frontmatter(body):
        for frontmatter in yaml.load_all(body, Loader=yaml.FullLoader):
            break
        if 'Date' in frontmatter:
            frontmatter['Date'] = datetime.fromisoformat(frontmatter['Date'])
        return frontmatter

    @staticmethod
    def _unique_id_from_body(body):
        return Issue._get_frontmatter(body)['ticket_id']

    @property
    def is_ticket_issue(self):
        try:
            self.unique_id
        except NonTicketIssueError:
            return False
        else:
            return True

    @property
    def unique_id(self):
        try:
            return self._unique_id_from_body(self._issue.body)
        except (yaml.YAMLError, KeyError):
            raise NonTicketIssueError(
                f"Issue {self.number} does not have a ticket ID"
            )

    @property
    def date_created(self):
        return self._issue.created_at

    @property
    def date_last_assigned(self):
        assigns = [
            event for event in self._issue.get_timeline()
            if event.event == "assigned"
        ]
        last_assignment = sorted(assigns, key=lambda x: x.created_at)[-1]
        return last_assignment.created_at

    @property
    def number(self):
        return self._issue.number

    @property
    def html_url(self):
        return self._issue.html_url

    @property
    def title(self):
        return self._issue.title

    @property
    def assignees(self):
        return [assignee.login for assignee in self._issue.assignees]


class NoIssue(Issue):
    """
    Special case to provide an :class:`.Issue` for dry runs.
    """
    def __init__(self):
        pass

    @property
    def unique_id(self):
        return "<UNKNOWN ID>"

    @property
    def number(self):
        return 0

    @property
    def html_url(self):
        return "<UNKNOWN URL>"

    @property
    def title(self):
        return "<UNKNOWN TITLE>"

    @property
    def assignees(self):
        return []



from .reminder_task import ReminderTask

from ..utils.datafiles import text_template

class UnassignedReminder(ReminderTask):
    REMINDER_TYPE = "Unassigned issue"
    DEFAULT_TEMPLATE_FILE = text_template("unassigned.txt")
    CONFIG = "unassigned-reminder"

    def _get_relevant_issues(self):
        # unassigned issues
        return self.bot.get_unassigned_issues()

    def _extract_date(self, issue, config):
        # issue creation date
        return issue.date_created


if __name__ == "__main__":
    UnassignedReminder.run_cli()



import os
import collections
import contextlib

import imaplib
import email
from email.utils import parsedate_to_datetime
from email.header import decode_header
from email.contentmanager import raw_data_manager

__all__ = ["Message", "Inbox"]

class Message:
    def __init__(self, extra, contents):
        self._extra = extra
        self._msg = email.message_from_bytes(contents)

    @property
    def unique_id(self):
        return self._msg["Message-ID"]

    @property
    def date(self):
        return parsedate_to_datetime(self._msg["Date"])

    @property
    def subject(self):
        output, encoding = decode_header(self._msg['subject'])[0]
        if isinstance(output, bytes):
            output = output.decode(encoding)

        return output

    def get(self, key):
        return self._msg.get(key)

    def _group_messages_by_content_type(self):
        messages = collections.defaultdict(list)
        for m in self._msg.walk():
            messages[m.get_content_type()].append(m)

        return dict(messages)

    @staticmethod
    def _get_desired_message(messages_by_content_type, type_order):
        for content_type in type_order:
            desired_messages = messages_by_content_type.get(content_type)
            if desired_messages:
                break

        return desired_messages[0]  # for now take the first one? Don't know

    def get_content(self, content_type_order=None):
        if content_type_order is None:
            content_type_order = ["text/plain", "text/html"]

        messages_by_content_type = self._group_messages_by_content_type()
        message = self._get_desired_message(messages_by_content_type,
                                            content_type_order)

        return raw_data_manager.get_content(message)


class Inbox:
    MESSAGE_CLASS = Message
    FETCH_STR = "RFC822"
    TYPE = "imap"

    def __init__(
        self,
        host,
        user,
        secret,
        mailbox="INBOX",
        ssl_port=993,
    ):
        self.host = host
        self.user = user
        self.secret = secret
        self.mailbox = mailbox
        self.ssl_port = ssl_port
        self.progress = lambda x: x

    def __repr__(self):
        return (f"{self.__class__.__name__}('{self.user}:"
                f"{self.ssl_port}/{self.mailbox}')")

    @classmethod
    def from_config(cls, config):
        kwargs = {k: v for k, v in config.items() if k != "type"}
        return cls(**kwargs)

    @contextlib.contextmanager
    def connection(self):
        """Get a connection to the IMAP server (use as context manager)."""
        password = os.environ.get(self.secret)
        imap = imaplib.IMAP4_SSL(self.host, port=self.ssl_port)
        imap.login(self.user, password)
        imap.select(self.mailbox)
        yield imap
        imap.logout()

    def _create_message(self, fetched):
        return self.MESSAGE_CLASS(*fetched[1][0])

    def _get_emails(self, search_string="ALL"):
        with self.connection() as imap:
            typ, data = imap.search(None, search_string)
            fetch_msgs = [
                imap.fetch(num, self.FETCH_STR)
                for num in self.progress(data[0].split())
            ]

        msgs = [self._create_message(m) for m in fetch_msgs]
        return msgs

    def get_emails(self, since=None):
        if since is None:
            search_string = "ALL"
        else:
            since_date = since.strftime("%d-%b-%Y")
            search_string = f"(SINCE {since_date})"
        return self._get_emails(search_string)

    def get_email(self, unique_id):
        for email in self.get_emails():
            if email.unique_id == unique_id:
                return email



from dataclasses import dataclass

@dataclass
class TeamMember:
    github: str
    email: str
    label: str  # label is gmail-specific

    @classmethod
    def from_config(cls, config):
        return cls(**config)



# Gmail-specific extensions
import shlex
import re

from .inbox import Message, Inbox

__all__ = ["GMailInbox", "GMessage"]

LABEL_PATTERN = re.compile(".*X-GM-LABELS \((?P<labels>[^)]*)\)")
MSGID_PATTERN = re.compile(".*X-GM-MSGID (?P<msgid>[0-9]+)")


def extract_from_pattern(string, pattern, pattern_name):
    if isinstance(string, bytes):
        string = str(string, 'utf-8')

    if match := pattern.match(string):
        found = match.group(pattern_name)
    else:
        found = ""

    return found


class GMessage(Message):
    @property
    def unique_id(self):
        return extract_from_pattern(self._extra, MSGID_PATTERN, "msgid")

    @property
    def labels(self):
        label_str = extract_from_pattern(self._extra, LABEL_PATTERN,
                                         "labels")
        return shlex.split(label_str)


class GMailInbox(Inbox):
    FETCH_STR = "(RFC822 X-GM-LABELS X-GM-MSGID)"
    MESSAGE_CLASS = GMessage
    TYPE = "gmail"

    def _toggle_labels(self, gm_msg_id, labels, direction):
        if direction not in ("+", "-"):
            raise ValueError(f"direction must be '+' or '-', not "
                             f"{direction}")

        with self.connection() as imap:
            labels_arg = f"({' '.join(labels)})"
            typ, data = imap.uid("SEARCH", f"X-GM-MSGID {gm_msg_id}")
            num = str(data[0], 'utf-8')
            store_args = f"{num} {direction}X-GM-LABELS {labels_arg}"
            imap.uid("STORE", store_args)

    def get_email(self, unique_id):
        msgs = self._get_emails(search_string=f"X-GM-MSGID {unique_id}")
        if len(msgs) != 1:
            raise RuntimeError("More than 1 message for unique ID "
                               f"'{unique_id}'")
        return msgs[0]

    def _add_labels(self, gm_msg_id, labels):
        return _toggle_labels(gm_msg_id, labels, "+")

    def _remove_labels(self, gm_msg_id, labels):
        return _toggle_labels(gm_msg_id, labels, "-")

    def set_labels(self, gm_msg_id, labels):
        msg = self.get_email(gm_msg_id)
        to_remove = set(msg.labels) - set(labels)
        to_add = set(labels) - set(msg.labels)

        self._add_labels(gm_msg_id, to_add)
        self._remove_labels(gm_msg_id, to_remove)



import os
import warnings

import github

from .issues import Issue, NonTicketIssueError

__all__ = ["SMTP", "Bot"]

class SMTP:
    def __init__(self, user, host, secret, port=465):
        self.user = user
        self.host = host
        self.secret = secret
        self.port = port

    def sendmail(self, email, recipients):
        with smtplib.SMTP_SSL(self.host, self.port) as smtp:
            smtp.login(user, os.environ.get(self.secret))
            smtp.sendmail(self.user, recipients, email.as_string())

    def __repr__(self):
        return (f"{self.__class__.__name__}('{self.user}:"
                f"{self.port}")


class Bot:
    def __init__(self, token_secret, repo, smtp=None):
        self.token_secret = token_secret
        self.reponame = repo
        self._github = None
        self.smtp = smtp

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.reponame}', {self.smtp})"

    @classmethod
    def from_config(cls, config):
        if smtp_config := config.get('sendmail'):
            smtp = SMTP(**smtp_config)
        else:
            smtp = None

        return cls(token_secret=config['token_name'],
                   repo=config['repo'],
                   smtp=smtp)

    @property
    def repo(self):
        if self._github is None:
            self._github = github.Github(os.environ[self.token_secret])
        return self._github.get_repo(self.reponame)

    def create_issue(self, title, content):
        return Issue(self.repo.create_issue(title, content))

    def get_issue(self, issue_num):
        return Issue(self.repo.get_issue(issue_num))

    def make_comment(self, issue_num, content):
        issue = self.repo.get_issue(issue_num)
        issue.create_comment(content)

    def get_unassigned_issues(self):
        return (
            Issue(iss)
            for iss in self.repo.get_issues(state="open")
            if not iss.assignees
        )

    def get_open_issues(self):
        return (Issue(iss) for iss in self.repo.get_issues(state="open"))

    def get_all_issues(self):
        return (Issue(iss) for iss in self.repo.get_issues(state="all"))

    def get_all_email_ticket_issues(self, nonticket="warn"):
        for iss in self.get_all_issues():
            try:
                iss.unique_id
            except NonTicketIssueError as e:
                if nonticket == "warn":
                    warnings.warn(str(e))
                elif nonticket == "error":
                    raise
            else:
                yield iss



import pathlib
import string

import yaml

from .utils.datafiles import workflow_template, DEFAULT_BUILD_CONFIG

import logging
_logger = logging.getLogger(__name__)


def get_workflow_template(template_name):
    path = pathlib.Path(template_name)
    if not path.exists():
        path = workflow_template(template_name)
        if not path.exists():
            raise RuntimeError("Unable to find workflow template "
                               f"{template_name}.")

    with open(path) as f:
        template = string.Template(f.read())

    return template


def build_substitutions(ticgithub_dict, builder_dict, config):
    dry = " --dry" if ticgithub_dict.get("force-dry") else ""
    if suffix := ticgithub_dict.get("suffix", ""):
        suffix = "-" + suffix

    default_install = "python -m pip install ticgithub"
    install = ticgithub_dict.get("install", default_install)

    name = builder_dict['config-name']

    ticgithub_sub = {
        "INSTALL_CMD": install,
        "SUFFIX": suffix,
        "DRY": dry,
    }

    workflow_builder_sub = {
        "NAME": name,
        "RUN_CMD": builder_dict['run-command'],
    }

    secrets = {
        "PAT": config['config']['bot']['token_name'],
        "INBOX": config['config']['inbox']['secret'],
    }
    sendmail_cfg = config['config']['bot'].get('sendmail', {})
    if sendmail_secret := sendmail_cfg.get('secret'):
        secrets["SENDMAIL"] = sendmail_secret

    build_params = builder_dict.get('build-params', {})
    workflow_build_params_sub = {
        value: config['workflows'][name][key]
        for key, value in build_params.items()
    }

    substitutions = dict(**ticgithub_sub, **workflow_builder_sub, **secrets,
                         **workflow_build_params_sub)
    return substitutions

def build_workflow(ticgithub_dict, builder_dict, config, dry):
    template = get_workflow_template(builder_dict['template'])
    substitutions = build_substitutions(ticgithub_dict, builder_dict,
                                        config)
    filename = substitutions["NAME"] + substitutions["SUFFIX"] + ".yml"
    directory = pathlib.Path(ticgithub_dict['install-path'])
    contents = template.substitute(substitutions)
    _logger.info(f"WRITING TO: {directory / filename}")
    _logger.info(f"Contents:\n{contents}")
    if not dry:
        directory.mkdir(parents=True, exist_ok=True)
        with open(directory / filename, mode="w") as f:
            f.write(contents)

def build(build_config, config, dry):
    ticgithub_dict = build_config.get('ticgithub', {})
    name_to_build_config = {builder['config-name']: builder
                            for builder in build_config['builders']}
    for workflow in config['workflows']:
        builder_dict = name_to_build_config[workflow]
        build_workflow(ticgithub_dict, builder_dict, config, dry)


def make_parser():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--dry', action="store_true", default=False)
    parser.add_argument('-c', '--config', type=str,
                        default=".ticgithub.yml")
    parser.add_argument('--build-config', type=str, default=None)
    parser.add_argument('--loglevel', type=str, default="WARNING")
    return parser


if __name__ == "__main__":
    parser = make_parser()
    opts = parser.parse_args()
    logging.basicConfig(level=getattr(logging, opts.loglevel))
    if not (build_config_file := opts.build_config):
        build_config_file = DEFAULT_BUILD_CONFIG
    with open(build_config_file) as f:
        build_config = yaml.load(f, Loader=yaml.FullLoader)

    with open(opts.config) as f:
        config = yaml.load(f, Loader=yaml.FullLoader)

    build(build_config, config, opts.dry)



import logging
_logger = logging.getLogger("__name__")

import yaml

from ..inbox import Inbox
from ..gmail import GMailInbox
from ..bot import Bot
from ..team import TeamMember

BOXTYPE_DISPATCH = {
    'gmail': GMailInbox
}

class Task:
    CONFIG = None
    def __init__(self, inbox, bot, team, config):
        self.inbox = inbox
        self.bot = bot
        self.team = team
        self.config = config

    @staticmethod
    def make_parser():
        _logger.debug("building parser")
        import argparse
        parser = argparse.ArgumentParser()
        parser.add_argument('--dry', action="store_true", default=False)
        parser.add_argument('-c', '--config', type=str,
                            default=".ticgithub.yml")
        parser.add_argument('--loglevel', type=str, default="INFO")
        return parser

    @classmethod
    def inject_config_parameters(cls, config, opts):
        return config, opts

    @staticmethod
    def use_parser(parser, args=None):
        _logger.debug("using parser")
        opts = parser.parse_args(args)
        logging.basicConfig(level=getattr(logging, opts.loglevel))
        with open(opts.config, 'r') as f:
            config = yaml.load(f, Loader=yaml.FullLoader)

        return config, opts

    @classmethod
    def run_cli(cls):
        parser = cls.make_parser()
        config, opts = cls.use_parser(parser)
        config, opts = cls.inject_config_parameters(config, opts)
        task = cls.from_config(config)
        task(opts.dry)

    @classmethod
    def from_config(cls, cfg_dict):
        _logger.debug(f"Building {cls} from config")
        config = cfg_dict['config']
        inbox_cfg = config['inbox']
        boxtype = BOXTYPE_DISPATCH[inbox_cfg['type']]
        inbox = boxtype.from_config(inbox_cfg)
        bot = Bot.from_config(config['bot'])
        team = [
            TeamMember.from_config(member)
            for member in config['team']
        ]
        config = cfg_dict['workflows'].get(cls.CONFIG)
        return cls(inbox, bot, team, config)

    def _build_config(self):
        """
        Convert the values of the config dict to usable Python objects.

        This takes the configuration values of the YAML config file, stored
        as basic Python types in ``self.config``, and returns a dictionary
        that is more straightforwar to use in `:method:._run`.
        """
        raise NotImplementedError()

    def _run(self, config, dry):
        raise NotImplementedError()

    def __call__(self, dry=False):
        # skip early if not in config or not active
        if not self.config or not self.config.get("active", True):
            _logger.info(f"EXITING: Config for '{self.config}' not active")
            return

        # calling param takes precedence (testing), then
        dry = dry or self.config.get('dry', False)

        cfg_dict = {k: v for k, v in self.config.items()
                    if k not in {"active", "dry"}}
        cfg = self._build_config()
        _logger.info(f"Running {self} with config {cfg}")
        self._run(cfg, dry)
